from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestSeed(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def test_request_seed_success(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x27\x05")
		self.conn.fromuserqueue.put(b"\x67\x05\x99\x88\x77\x66")	# Positive response

	def _test_request_seed_success(self):
		response = self.udsclient.request_seed(0x05)
		self.assertEqual(response.service_data.seed, b"\x99\x88\x77\x66")

	def test_request_seed_success_spr(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x27\x85")
		self.conn.fromuserqueue.put('wait')	# Synchronize 

	def _test_request_seed_success_spr(self):
		with self.udsclient.suppress_positive_response:
			response = self.udsclient.request_seed(0x05)
			self.assertEqual(response, None)
		self.conn.fromuserqueue.get(timeout=0.2)	#Avoid closing connection prematurely

	def test_request_seed_denied_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x22")	# Conditions Not Correct

	def _test_request_seed_denied_exception(self):
		with self.assertRaises(NegativeResponseException) as handle:
			self.udsclient.request_seed(0x05)
		response = handle.exception.response

		self.assertTrue(response.valid)
		self.assertFalse(response.positive)
		self.assertTrue(issubclass(response.service, services.SecurityAccess))
		self.assertEqual(response.code, 0x22)

	def test_request_seed_denied_no_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x22")	# Conditions Not Correct

	def _test_request_seed_denied_no_exception(self):
		self.udsclient.config['exception_on_negative_response'] = False
		response = self.udsclient.request_seed(0x05)

		self.assertTrue(response.valid)
		self.assertFalse(response.positive)
		self.assertTrue(issubclass(response.service, services.SecurityAccess))
		self.assertEqual(response.code, 0x22)

	def test_request_seed_bad_subfn_exception(self):
		self.wait_request_and_respond(b"\x67\x06\x99\x88\x77\x66")	# Positive response with wrong subfunction

	def _test_request_seed_bad_subfn_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.request_seed(0x05)

	def test_request_seed_bad_subfn_no_exception(self):
		self.wait_request_and_respond(b"\x67\x06\x99\x88\x77\x66")	# Positive response with wrong subfunction

	def _test_request_seed_bad_subfn_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.request_seed(0x05)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_request_seed_incomplete_response_exception(self):
		self.wait_request_and_respond(b"\x67\x05")	# Positive response with no seed

	def _test_request_seed_incomplete_response_exception(self):
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.request_seed(0x05)

	def test_request_seed_incomplete_response_no_exception(self):
		self.wait_request_and_respond(b"\x67\x05")	# Positive response with no seed

	def _test_request_seed_incomplete_response_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.request_seed(0x05)
		self.assertFalse(response.valid)

	def test_request_seed_invalidservice_exception(self):
		self.wait_request_and_respond(b"\x00\x05") #Inexistent Service

	def _test_request_seed_invalidservice_exception(self):
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.request_seed(0x05)

	def test_request_seed_invalidservice_no_exception(self):
		self.wait_request_and_respond(b"\x00\x05") #Inexistent Service

	def _test_request_seed_invalidservice_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.request_seed(0x05)
		self.assertFalse(response.valid)

	def test_request_seed_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_request_seed_wrongservice_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.request_seed(0x05)

	def test_request_seed_wrongservice_no_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_request_seed_wrongservice_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.request_seed(0x05)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)
		
	def test_request_seed_bad_param(self):
		pass

	def _test_request_seed_bad_param(self):
		with self.assertRaises(ValueError):
			self.udsclient.request_seed(0x80)

		with self.assertRaises(ValueError):
			self.udsclient.request_seed(-1)


class TestSendKey(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def test_send_key_success(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x27\x06\x11\x22\x33\x44")
		self.conn.fromuserqueue.put(b"\x67\x06")	# Positive response

	def _test_send_key_success(self):
		response = self.udsclient.send_key(0x06,b"\x11\x22\x33\x44")
		self.assertTrue(response.positive)

	def test_send_key_denied_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x35")	# InvalidKey

	def _test_send_key_denied_exception(self):
		with self.assertRaises(NegativeResponseException) as handle:
			self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
		response = handle.exception.response

		self.assertTrue(response.valid)
		self.assertFalse(response.positive)
		self.assertTrue(issubclass(response.service, services.SecurityAccess))
		self.assertEqual(response.code, 0x35)

	def test_send_key_denied_no_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x35")	# InvalidKey

	def _test_send_key_denied_no_exception(self):
		self.udsclient.config['exception_on_negative_response'] = False
		response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

		self.assertTrue(response.valid)
		self.assertFalse(response.positive)
		self.assertTrue(issubclass(response.service, services.SecurityAccess))
		self.assertEqual(response.code, 0x35)

	def test_send_key_bad_subfn_exception(self):
		self.wait_request_and_respond(b"\x67\x08\x99\x88\x77\x66")	# Positive response with wrong subfunction

	def _test_send_key_bad_subfn_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

	def test_send_key_bad_subfn_no_exception(self):
		self.wait_request_and_respond(b"\x67\x08\x99\x88\x77\x66")	# Positive response with wrong subfunction

	def _test_send_key_bad_subfn_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_send_key_invalidservice_exception(self):
		self.wait_request_and_respond(b"\x00\x06") #Inexistent Service

	def _test_send_key_invalidservice_exception(self):
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

	def test_send_key_invalidservice_no_exception(self):
		self.wait_request_and_respond(b"\x00\x06") #Inexistent Service

	def _test_send_key_invalidservice_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
		self.assertFalse(response.valid)

	def test_send_key_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_send_key_wrongservice_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

	def test_send_key_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_send_key_wrongservice_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_send_key_bad_param(self):
		pass

	def _test_send_key_bad_param(self):
		with self.assertRaises(ValueError):
			self.udsclient.send_key(0x80, b"\x11\x22\x33\x44")

		with self.assertRaises(ValueError):
			self.udsclient.send_key(-1, b"\x11\x22\x33\x44")

		with self.assertRaises(ValueError):
			self.udsclient.send_key(1, 1)

		with self.assertRaises(ValueError):
			self.udsclient.send_key(1, 'xxx')


class TestUnlockSecurityService(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def dummy_algo(self, seed, params=None):
		key = bytearray(seed)
		for i in range(len(key)):
			key[i] ^= params
		return bytes(key)

	def test_unlock_success(self):
		for i in range(2):
			request = self.conn.touserqueue.get(timeout=0.2)
			self.assertEqual(request, b"\x27\x07")		# Request seed
			self.conn.fromuserqueue.put(b"\x67\x07\x11\x22\x33\x44")	# Positive response
			request = self.conn.touserqueue.get(timeout=0.2)
			self.assertEqual(request, b"\x27\x08\xEE\xDD\xCC\xBB")
			self.conn.fromuserqueue.put(b"\x67\x08")	# Positive response

	def _test_unlock_success(self):
		self.udsclient.config['security_algo'] = self.dummy_algo
		self.udsclient.config['security_algo_params'] = 0xFF
		response = self.udsclient.unlock_security_access(0x07)	
		response = self.udsclient.unlock_security_access(0x08)	
		self.assertTrue(response.positive)


	def test_unlock_seed_fail_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x11")	

	def _test_unlock_seed_fail_exception(self):
		self.udsclient.config['security_algo'] = self.dummy_algo
		self.udsclient.config['security_algo_params'] = 0xFF
		with self.assertRaises(NegativeResponseException):
			response = self.udsclient.unlock_security_access(0x07)	

	def test_unlock_seed_fail_no_exception(self):
		self.wait_request_and_respond(b"\x7F\x27\x11")	

	def _test_unlock_seed_fail_no_exception(self):
		self.udsclient.config['security_algo'] = self.dummy_algo
		self.udsclient.config['security_algo_params'] = 0xFF
		self.udsclient.config['exception_on_negative_response'] = False
		response = self.udsclient.unlock_security_access(0x07)
		self.assertTrue(response.valid)
		self.assertFalse(response.positive)

	def test_no_algo_set(self):
		pass

	def _test_no_algo_set(self):
		with self.assertRaises(NotImplementedError):
			self.udsclient.unlock_security_access(0x07)


from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *
from udsoncan import DidCodec
import struct

from test.ClientServerTest import ClientServerTest


class StubbedDidCodec(DidCodec):
	def encode(self, did_value):
		return struct.pack('B', did_value+1)

	def decode(self, did_payload):
		return struct.unpack('B', did_payload)[0] - 1

	def __len__(self):
		return 1

class TestReadDataByIdentifier(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def postClientSetUp(self):
		self.udsclient.config["data_identifiers"] = {
			1 : '>H',
			2 : '<H',
			3 : StubbedDidCodec
		}

	def test_wdbi_single_success1(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2E\x00\x01\x12\x34")
		self.conn.fromuserqueue.put(b"\x6E\x00\x01")	# Positive response

	def _test_wdbi_single_success1(self):
		response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
		self.assertTrue(response.positive)

	def test_wdbi_single_success1_spr_no_effect(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2E\x00\x01\x12\x34")
		self.conn.fromuserqueue.put(b"\x6E\x00\x01")	# Positive response

	def _test_wdbi_single_success1_spr_no_effect(self):
		with self.udsclient.suppress_positive_response:
			response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
			self.assertTrue(response.positive)

	def test_wdbi_single_success2(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2E\x00\x02\x34\x12")
		self.conn.fromuserqueue.put(b"\x6E\x00\x02")	# Positive response

	def _test_wdbi_single_success2(self):
		self.udsclient.write_data_by_identifier(did = 2, value=0x1234)

	def test_wdbi_incomplete_response_exception(self):
		self.wait_request_and_respond(b"\x6E\x00")	#Incomplete response

	def _test_wdbi_incomplete_response_exception(self):
		with self.assertRaises(InvalidResponseException):
			self.udsclient.write_data_by_identifier(did = 1, value=0x1234)

	def test_wdbi_incomplete_response_no_exception(self):
		self.wait_request_and_respond(b"\x6E\x00")	#Incomplete response

	def _test_wdbi_incomplete_response_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
		self.assertFalse(response.valid)

	def test_wdbi_unknown_did_exception(self):
		self.wait_request_and_respond(b"\x6E\x00\x09")	# Positive response

	def _test_wdbi_unknown_did_exception(self):
		with self.assertRaises(UnexpectedResponseException):
			self.udsclient.write_data_by_identifier(did = 1, value=0x1234)			
	
	def test_wdbi_unknown_did_no_exception(self):
		self.wait_request_and_respond(b"\x6E\x00\x09")	# Positive response

	def _test_wdbi_unknown_did_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)			
	
	def test_wdbi_unwanted_did_exception(self):
		self.wait_request_and_respond(b"\x6E\x00\x02")	# Positive response

	def _test_wdbi_unwanted_did_exception(self):
		with self.assertRaises(UnexpectedResponseException):
			self.udsclient.write_data_by_identifier(did = 1, value=0x1234)			
	
	def test_wdbi_unwanted_did_no_exception(self):
		self.wait_request_and_respond(b"\x6E\x00\x02")	# Positive response

	def _test_wdbi_unwanted_did_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_wdbi_invalidservice_exception(self):
		self.wait_request_and_respond(b"\x00\x00\x01")	# Service is inexistant

	def _test_wdbi_invalidservice_exception(self):
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.write_data_by_identifier(did=1, value=0x1234)

	def test_wdbi_invalidservice_no_exception(self):
		self.wait_request_and_respond(b"\x00\x00\x01")	# Service is inexistant

	def _test_wdbi_invalidservice_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.write_data_by_identifier(did=1, value=0x1234)
		self.assertFalse(response.valid)

	def test_wdbi_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x50\x00\x01")	# Valid service, but not the one requested

	def _test_wdbi_wrongservice_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.write_data_by_identifier(did=1, value=0x1234)

	def test_wdbi_wrongservice_no_exception(self):
		self.wait_request_and_respond(b"\x50\x00\x01")	# Valid service, but not the one requested

	def _test_wdbi_wrongservice_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_no_config(self):
		pass

	def _test_no_config(self):
		with self.assertRaises(ConfigError):
			self.udsclient.write_data_by_identifier(did = 4, value=0x1234) 


from . import *
from udsoncan.Response import Response
from udsoncan.exceptions import *

class ReadMemoryByAddress(BaseService):
	_sid = 0x23
	_use_subfunction = False

	supported_negative_response = [	 Response.Code.IncorrectMessageLegthOrInvalidFormat,
							Response.Code.ConditionsNotCorrect,
							Response.Code.RequestOutOfRange,
							Response.Code.SecurityAccessDenied
							]

	@classmethod
	def make_request(cls, memory_location):
		"""
		Generates a request for ReadMemoryByAddress

		:param memory_location: The address and the size of the memory block to read.
		:type memory_location: :ref:`MemoryLocation <MemoryLocation>`

		:raises ValueError: If parameters are out of range, missing or wrong type
		"""		
		from udsoncan import Request, MemoryLocation

		if not isinstance(memory_location, MemoryLocation):
			raise ValueError('Given memory location must be an instance of MemoryLocation')

		request =  Request(service=cls)
		request.data = b''
		request.data += memory_location.alfid.get_byte() # AddressAndLengthFormatIdentifier
		request.data += memory_location.get_address_bytes()
		request.data += memory_location.get_memorysize_bytes()

		return request

	@classmethod
	def interpret_response(cls, response):
		"""
		Populates the response ``service_data`` property with an instance of :class:`ReadMemoryByAddress.ResponseData<udsoncan.services.ReadMemoryByAddress.ResponseData>`

		:param response: The received response to interpret
		:type response: :ref:`Response<Response>`

		:raises InvalidResponseException: If length of ``response.data`` is too short
		"""		
		if len(response.data) < 1: 	
			raise InvalidResponseException(response, "Response data must be at least 1 byte")

		response.service_data = cls.ResponseData()
		response.service_data.memory_block = response.data

	class ResponseData(BaseResponseData):
		"""
		.. data:: memory_block

			bytes object reflecting the content of the read memory
		"""
		def __init__(self):
			super().__init__(ReadMemoryByAddress)
			self.memory_block = None


from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestClearDtc(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def test_clear_dtc_success(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x14\x12\x34\x56")
		self.conn.fromuserqueue.put(b"\x54")	# Positive response

	def _test_clear_dtc_success(self):
		response = self.udsclient.clear_dtc(0x123456)
		self.assertTrue(response.valid)
		self.assertTrue(response.positive)

	def test_clear_dtc_spr_no_effect(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x14\x12\x34\x56")
		self.conn.fromuserqueue.put(b"\x54")	# Positive response

	def _test_clear_dtc_spr_no_effect(self):
		with self.udsclient.suppress_positive_response:
			response = self.udsclient.clear_dtc(0x123456)
			self.assertTrue(response.valid)
			self.assertTrue(response.positive)

	def test_clear_dtc_all(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x14\xFF\xFF\xFF")
		self.conn.fromuserqueue.put(b"\x54")	# Positive response

	def _test_clear_dtc_all(self):
		response = self.udsclient.clear_dtc()
		self.assertTrue(response.valid)
		self.assertTrue(response.positive)

	def test_clear_dtc_denied_exception(self):
		self.wait_request_and_respond(b"\x7F\x14\x31") #Request Out Of Range

	def _test_clear_dtc_denied_exception(self):
		with self.assertRaises(NegativeResponseException) as handle:
			self.udsclient.clear_dtc(0x123456)

	def test_clear_dtc_denied_no_exception(self):
		self.wait_request_and_respond(b"\x7F\x14\x31") #Request Out Of Range

	def _test_clear_dtc_denied_no_exception(self):
		self.udsclient.config['exception_on_negative_response'] = False
		response =self.udsclient.clear_dtc(0x123456)
		self.assertTrue(response.valid)
		self.assertFalse(response.positive)
		self.assertEqual(response.code, 0x31)

	def test_clear_dtc_invalidservice_exception(self):
		self.wait_request_and_respond(b"\x00") #Inexistent Service

	def _test_clear_dtc_invalidservice_exception(self):
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.clear_dtc(0x123456)

	def test_clear_dtc_invalidservice_no_exception(self):
		self.wait_request_and_respond(b"\x00") #Inexistent Service

	def _test_clear_dtc_invalidservice_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.clear_dtc(0x123456)
		self.assertFalse(response.valid)

	def test_clear_dtc_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_clear_dtc_wrongservice_exception(self):
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.clear_dtc(0x123456)

	def test_clear_dtc_wrongservice_no_exception(self):
		self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

	def _test_clear_dtc_wrongservice_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.clear_dtc(0x123456)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_bad_param(self):
		pass

	def _test_bad_param(self):
		with self.assertRaises(ValueError):
			self.udsclient.clear_dtc(0x1000000)

		with self.assertRaises(ValueError):
			self.udsclient.clear_dtc(-1)


from udsoncan import DataIdentifier, Routine, Units
from test.UdsTest import UdsTest

class TestDefinitions(UdsTest):
	def test_data_identifier_name_from_id(self):
		for i in range(0x10000):
			name = DataIdentifier.name_from_id(i)
			self.assertTrue(isinstance(name, str))
	
	def test_routine_name_from_id(self):
		for i in range(0x10000):
			name = Routine.name_from_id(i)
			self.assertTrue(isinstance(name, str))
		


from . import *
from udsoncan.Response import Response
from udsoncan.exceptions import *

class ControlDTCSetting(BaseService):
	_sid = 0x85

	supported_negative_response = [	Response.Code.SubFunctionNotSupported, 
							Response.Code.IncorrectMessageLegthOrInvalidFormat,
							Response.Code.ConditionsNotCorrect,
							Response.Code.RequestOutOfRange
							]
	class SettingType(BaseSubfunction):
		"""
		ControlDTCSetting defined subfunctions
		"""

		__pretty_name__ = 'setting type'

		on = 1
		off = 2
		vehicleManufacturerSpecific = (0x40, 0x5F)	# To be able to print textual name for logging only.
		systemSupplierSpecific = (0x60, 0x7E)		# To be able to print textual name for logging only.

	@classmethod
	def make_request(cls, setting_type, data = None):
		"""
		Generates a request for ControlDTCSetting

		:param setting_type: Service subfunction. Allowed values are from 0 to 0x7F
		:type setting_type: int

		:param data: Optional additional data sent with the request called `DTCSettingControlOptionRecord`
		:type data: bytes

		:raises ValueError: If parameters are out of range, missing or wrong type
		"""		
		from udsoncan import Request

		ServiceHelper.validate_int(setting_type, min=0, max=0x7F, name='Setting type')
		if data is not None:
			if not isinstance(data, bytes):
				raise ValueError('data must be a valid bytes object')

		return Request(service=cls, subfunction=setting_type, data=data)

	@classmethod
	def interpret_response(cls, response):
		"""
		Populates the response ``service_data`` property with an instance of :class:`ControlDTCSetting.ResponseData<udsoncan.services.ControlDTCSetting.ResponseData>`

		:param response: The received response to interpret
		:type response: :ref:`Response<Response>`

		:raises InvalidResponseException: If length of ``response.data`` is too short
		"""		
		if len(response.data) < 1: 	
			raise InvalidResponseException(response, "Response data must be at least 1 byte")

		response.service_data = cls.ResponseData()
		response.service_data.setting_type_echo = response.data[0]

	class ResponseData(BaseResponseData):
		"""
		.. data:: setting_type_echo

			Request subfunction echoed back by the server
		"""		
		def __init__(self):
			super().__init__(ControlDTCSetting)
			self.setting_type_echo = None



from udsoncan.client import Client
from udsoncan import services, MemoryLocation, DataFormatIdentifier
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestUpload(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def test_request_upload_success(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x35\x00\x12\x12\x34\xFF")
		self.conn.fromuserqueue.put(b"\x75\x20\xab\xcd")	# Positive response

	def _test_request_upload_success(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertEqual(response.service_data.max_length,0xabcd)

	def test_request_upload_success_spr_no_effect(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x35\x00\x12\x12\x34\xFF")
		self.conn.fromuserqueue.put(b"\x75\x20\xab\xcd")	# Positive response

	def _test_request_upload_success_spr_no_effect(self):
		with self.udsclient.suppress_positive_response:
			memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
			response = self.udsclient.request_upload(memory_location=memloc)
			self.assertEqual(response.service_data.max_length,0xabcd)

	def test_request_upload_config_format(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x35\x00\x24\x00\x00\x12\x34\x00\xFF")	# dfi = 24 and 0 padding
		self.conn.fromuserqueue.put(b"\x75\x20\xab\xcd")	# Positive response

	def _test_request_upload_config_format(self):
		self.udsclient.config = {'server_address_format':32, 'server_memorysize_format':16}
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertEqual(response.service_data.max_length,0xabcd)

	def test_request_upload_success_lfid(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x35\x00\x12\x12\x34\xFF")
		self.conn.fromuserqueue.put(b"\x75\x23\xab\xcd")	# Positive response

	def _test_request_upload_success_lfid(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertEqual(response.service_data.max_length,0xabcd)

	def test_request_upload_success_dfi(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x35\x52\x12\x12\x34\xFF")
		self.conn.fromuserqueue.put(b"\x75\x20\xab\xcd")	# Positive response

	def _test_request_upload_success_dfi(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		dfi  =DataFormatIdentifier(compression=5, encryption=2)
		response = self.udsclient.request_upload(memory_location=memloc, dfi=dfi)
		self.assertEqual(response.service_data.max_length,0xabcd)

	def test_incomplete_nblock_response_exception(self):
		self.wait_request_and_respond(b"\x75\x40\xab\xcd")	# Missing 2 bytes to complete number of block

	def _test_incomplete_nblock_response_exception(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		with self.assertRaises(InvalidResponseException):
			self.udsclient.request_upload(memory_location=memloc)

	def test_incomplete_nblock_response_no_exception(self):
		self.wait_request_and_respond(b"\x75\x40\xab\xcd")	# Missing 2 bytes to complete number of block

	def _test_incomplete_nblock_response_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False	
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertFalse(response.valid)

	def test_request_upload_invalidservice_exception(self):
		self.wait_request_and_respond(b"\x00\x20\x12\x34") #Inexistent Service

	def _test_request_upload_invalidservice_exception(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		with self.assertRaises(InvalidResponseException) as handle:
			self.udsclient.request_upload(memory_location=memloc)

	def test_request_upload_invalidservice_no_exception(self):
		self.wait_request_and_respond(b"\x00\x20\x12\x34") #Inexistent Service

	def _test_request_upload_invalidservice_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False	
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertFalse(response.valid)

	def test_request_upload_wrongservice_exception(self):
		self.wait_request_and_respond(b"\x7E\x20\x12\x34") # Valid but wrong service (Tester Present)

	def _test_request_upload_wrongservice_exception(self):
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		with self.assertRaises(UnexpectedResponseException) as handle:
			self.udsclient.request_upload(memory_location=memloc)

	def test_request_upload_wrongservice_no_exception(self):
		self.wait_request_and_respond(b"\x7E\x20\x12\x34") # Valid but wrong service (Tester Present)

	def _test_request_upload_wrongservice_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
		response = self.udsclient.request_upload(memory_location=memloc)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)

	def test_bad_params(self):
		pass

	def _test_bad_params(self):
		with self.assertRaises(ValueError) as handle:
			self.udsclient.request_upload(memory_location=1)
		
		with self.assertRaises(ValueError) as handle:
			self.udsclient.request_upload(memory_location="asd")



from . import *
from udsoncan.Response import Response
from udsoncan.exceptions import *
import struct


class ReadDataByIdentifier(BaseService):
	_sid = 0x22
	_use_subfunction = False


	supported_negative_response = [	 Response.Code.IncorrectMessageLegthOrInvalidFormat,
							Response.Code.ConditionsNotCorrect,
							Response.Code.RequestOutOfRange,
							Response.Code.SecurityAccessDenied
							]	
	
	@classmethod
	def validate_didlist_input(cls, dids):
		if not isinstance(dids, int) and not isinstance(dids, list):
			raise ValueError("Data Identifier must either be an integer or a list of integer")

		if isinstance(dids, int):
			ServiceHelper.validate_int(dids, min=0, max=0xFFFF, name='Data Identifier')
		
		if isinstance(dids, list):
			for did in dids:
				ServiceHelper.validate_int(did, min=0, max=0xFFFF, name='Data Identifier')


		return [dids] if not isinstance(dids, list) else dids

	@classmethod
	def make_request(cls, didlist, didconfig):
		"""
		Generates a request for ReadDataByIdentifier

		:param didlist: List of data identifier to read.
		:type didlist: list[int]

		:param didconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string 
		:type didconfig: dict[int] = :ref:`DidCodec<DidCodec>`

		:raises ValueError: If parameters are out of range, missing or wrong type
		:raises ConfigError: If didlist contains a DID not defined in didconfig
		"""		
		from udsoncan import Request
		didlist = cls.validate_didlist_input(didlist)

		req = Request(cls)
		ServiceHelper.check_did_config(didlist, didconfig)
		req.data = struct.pack('>'+'H'*len(didlist), *didlist) #Encode list of DID

		return req

	@classmethod
	def interpret_response(cls, response, didlist, didconfig, tolerate_zero_padding=True):
		"""
		Populates the response ``service_data`` property with an instance of :class:`ReadDataByIdentifier.ResponseData<udsoncan.services.ReadDataByIdentifier.ResponseData>`

		:param response: The received response to interpret
		:type response: :ref:`Response<Response>`
		
		:param didlist:  List of data identifiers used for the request.
		:type didlist: list[int]
		
		:param didconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string 
		:type didconfig: dict[int] = :ref:`DidCodec<DidCodec>`

		:param tolerate_zero_padding: Ignore trailing zeros in the response data avoiding raising false :class:`InvalidResponseException<udsoncan.exceptions.InvalidResponseException>`.
		:type tolerate_zero_padding: bool

		:raises ValueError: If parameters are out of range, missing or wrong type
		:raises ConfigError: If ``didlist`` parameter or response contains a DID not defined in ``didconfig``.
		:raises InvalidResponseException: If response data is incomplete or if DID data does not match codec length.
		"""	
		from udsoncan import DidCodec

		didlist = cls.validate_didlist_input(didlist)
		didconfig = ServiceHelper.check_did_config(didlist, didconfig)
		
		response.service_data = cls.ResponseData()
		response.service_data.values = {}

		# Parsing algorithm to extract DID value
		offset = 0
		while True:
			if len(response.data) <= offset:
				break	# Done

			if len(response.data) <= offset +1:
				if tolerate_zero_padding and response.data[-1] == 0:	# One extra byte, but it's a 0 and we accept that. So we're done
					break
				raise InvalidResponseException(response, "Response given by server is incomplete.")

			did = struct.unpack('>H', response.data[offset:offset+2])[0]	# Get the DID number
			if did == 0 and did not in didconfig and tolerate_zero_padding: # We read two zeros and that is not a DID bu we accept that. So we're done.
				if response.data[offset:] == b'\x00' * (len(response.data) - offset):
					break

			if did not in didconfig:	# Already checked in check_did_config. Paranoid check
				raise ConfigError(key=did, msg='Actual data identifier configuration contains no definition for data identifier 0x%04x' % did)
			
			codec = DidCodec.from_config(didconfig[did])
			offset+=2

			if len(response.data) < offset+len(codec):
				raise InvalidResponseException(response, "Value fo data identifier 0x%04x was incomplete according to definition in configuration" % did)

			subpayload = response.data[offset:offset+len(codec)]
			offset += len(codec)	# Codec must define a __len__ function that matches the encoded payload length.
			val = codec.decode(subpayload)
			response.service_data.values[did] = val

		return response

	class ResponseData(BaseResponseData):
		"""
		.. data:: values

			Dictionary mapping the DID (int) with the value returned by the associated :ref:`DidCodec<DidCodec>`.decode method
		"""				
		def __init__(self):
			super().__init__(ReadDataByIdentifier)

			self.values = None


import inspect
def service_name(service):
	if inspect.isclass(service):
		return service.__name__
	else:
		return service.__class__.__name__

class TimeoutException(Exception):
	"""
	Simple extension of ``Exception`` with no additional property. Raised when a timeout in the communication happens.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

class NegativeResponseException(Exception):
	"""
	Raised when the server returns a negative response (response code starting by 0x7F).
	The response that triggered the exception is available in ``e.response``

	:param response: The response that triggered the exception
	:type response: :ref:`Response<Response>`
	"""
	def __init__(self, response, *args, **kwargs):
		self.response = response
		msg = self.make_msg(response)
		if len(args) > 0 :
			msg += "\n"+args[0]
		super().__init__(msg, *args, **kwargs)

	def make_msg(self, response):
		
		return "%s service execution returned a negative response %s (0x%x)" % (response.service.get_name(), response.code_name, response.code)

class InvalidResponseException(Exception):
	"""
	Raised when a service fails to decode a server response data. A bad message length or a value that is out of range may both be valid causes.
	The response that triggered the exception is available in ``e.response``

	:param response: The response that triggered the exception
	:type response: :ref:`Response<Response>`
	"""

	def __init__(self, response, *args, **kwargs):
		self.response = response
		msg = self.make_msg(response)
		if len(args) > 0 :
			msg += "\n"+args[0]
		super().__init__(msg, *args, **kwargs)

	def make_msg(self, response):
		servicename = response.service.get_name() if response.service is not None else ""
		return "%s service execution returned an invalid response. Reason : %s" % (servicename, response.invalid_reason)

class UnexpectedResponseException(Exception):
	"""
	Raised when the client receives a valid response but considers the one received to not be the expected response.
	The response that triggered the exception is available in ``e.response``

	:param response: The response that triggered the exception
	:type response: :ref:`Response<Response>`

	:param details: Additional details about the error
	:type details: string
	"""
	def __init__(self, response, details="<No details given>", *args, **kwargs):
		self.response = response
		msg = self.make_msg(response, details)
		if len(args) > 0 :
			msg += "\n"+args[0]
		super().__init__(msg, *args, **kwargs)

	def make_msg(self, response, details):
		servicename = response.service.get_name() if response.service is not None else ""
		return "service execution returned a valid response for service %s, but unexpected. Details : %s " % (servicename, details)

class ConfigError(Exception):
	"""
	Raised when a bad configuration element is encountered.
	
	:param key: The configuration key that failed to resolve properly
	:type key: object

	"""
	def __init__(self, key, msg="<No details given>", *args, **kwargs):
		self.key=key
		super().__init__(msg, *args, **kwargs)
